#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
ECODOPS Project
Source code implementing the tsne algorithm for multi-D data visualization

WHO                 WHEN            WHAT
------------------  ----------      -------------------------------------------------------------
M. Brescia          29/09/2018      Created
M. D'Addona         08/01/2018      Removed redundant dependencies: pandas, configobj
M. D'Addona         08/01/2018      Added optional FITS support
M. Brescia          04/02/2020      bug fixed with output (now first column is the original class
									Introduced also new parameter n_jobs for MPI execution
-------------------------------------------------------------------------------------------------
"""

import time
import datetime
import os
import numpy as np
from configparser import ConfigParser
import argparse
from sklearn.manifold import TSNE
from sklearn import preprocessing
import matplotlib.pyplot as plt
import csv
import sys
from numpy.lib import recfunctions as rfn

try:
    from astropy.table import Table
except:
    HAS_ASTROPY = False
    print("WARNING: Astropy not installed, only CSV files are supported\n")
else:
    HAS_ASTROPY = True


def get_io(conf):
    """ Read the I/O section of the config file"""
    inputs = conf['I/O']['Input']
    output = conf['I/O']['Output']
    logfile_path = conf['I/O']['Log file'].lower()
    normalization = conf["I/O"]["Normalization"]
    return inputs, output, logfile_path, normalization


def get_params(conf):
    """ Read the Parameter section of the config file"""
    class_discrimination = conf["TSNE parameters"]["Class_Discrimination"]
    n_components = conf["TSNE parameters"]["n_components"]
    perplexity = conf["TSNE parameters"]["perplexity"]
    learning_rate = conf["TSNE parameters"]["learning_rate"]
    n_iter = conf["TSNE parameters"]["n_iter"]
    verbose = conf["TSNE parameters"]["verbose"]
    early_exaggeration = conf["TSNE parameters"]["early_exaggeration"]
    n_iter_without_progress = conf["TSNE parameters"]["n_iter_without_progress"]
    init = conf["TSNE parameters"]["init"]
    random_state = conf["TSNE parameters"]["random_state"]
    if random_state == "None":
        return str(class_discrimination), int(n_components), float(perplexity), float(learning_rate), int(n_iter), int(
            verbose), float(early_exaggeration), int(n_iter_without_progress), str(init), bool(random_state)
    else:
        return str(class_discrimination), int(n_components), float(perplexity), float(learning_rate), int(n_iter), int(
            verbose), float(early_exaggeration), int(n_iter_without_progress), str(init), int(random_state)


def load_csv(path):
    """Open the CSV, splits it and does the preprocessing"""
    with open(path) as csv_file:
        csv_reader = csv.reader(csv_file)
        ID = np.asarray(next(csv_reader))
        data = np.asarray([[x for x in r] for r in csv_reader], dtype=np.float32)
        X = data[:, :-1]
        Y = data[:, -1]
    return X, Y, ID


def load_ap(path):
    """Open input file using astropy Table bakcend"""
    t = Table.read(path)
    X = np.empty((len(t), len(t.colnames)-1))
    Y = np.empty(len(t))
    for n, col in enumerate(t.colnames[:-1]):
        X[:,n] = t[col].astype(np.float64)
    Y[:] = t[t.colnames[-1]].astype(np.float64)
    return X, Y, np.asarray(t.colnames)


def split_path(path):
    path = os.path.abspath(path)
    dirname = os.path.dirname(path)  # directory where the file is saved
    filename = os.path.basename(path)  # file name with extension
    # print(dirname, filename, path)
    extension = filename.split('.')[1]
    return dirname, filename, extension, path


def check_file(path):
    """Check if the Congif.ini exists"""
    dirname, filename, extension, path = split_path(path)
    if not filename in os.listdir(dirname):
        print('ERROR: file {} does not exist'.format(path))
        exit()
    return path


def check_outputdir(path, create_dir=True):
    """Check if the output directory exixts"""
    path = os.path.abspath(path)
    dirname = os.path.dirname(path)  # directory where the file is saved
    filename = os.path.basename(path)  # file name with extension
    if filename in os.listdir(dirname):
        print(
            'WARNING: directory {} already exist in {}\nname automatically change adding current date'.format(filename,
                                                                                                              dirname))
        lt = time.localtime(time.time())
        localClock = str(lt.tm_mon) + '_' + str(lt.tm_mday) + '_' + str(lt.tm_hour) + '_' + str(lt.tm_min) + '_' + str(
            lt.tm_sec)
        filename += '_' + localClock
        print('Actual directory name: {}'.format(filename))
        path = dirname + os.sep + filename

    os.makedirs(path)
    return path


def normalize_list(X):
    max_value = np.amax(X)
    min_value = np.amin(X)
    X = [(i - min_value) / (max_value - min_value) for i in X]
    return X

  # File where you need to keep the logs


class Logger(object):

    def __init__(self):
        self.terminal = sys.stdout
        self.log = open(output + os.sep + "logfile.log", "a")

    def write(self, message):
        self.terminal.write(message)
        self.log.write(message)

    def flush(self):
        # this flush method is needed for python 3 compatibility.
        # this handles the flush command by doing nothing.
        # you might want to specify some extra behavior here.
        pass


start_time = time.time()

if __name__ == "__main__":

    now = datetime.datetime.now()

    AP = argparse.ArgumentParser()
    AP.add_argument("-c", "--conf", default="Config.ini",
                    type=str, help="configuration file path")
    ARGS = vars(AP.parse_args())
    CONFIGFILE = ARGS["conf"]
    CONFIGFILE = check_file(CONFIGFILE)

    CONFIG = ConfigParser(inline_comment_prefixes='#')
    CONFIG.read(CONFIGFILE)

    inputs, output, logfile_path, normalization = get_io(CONFIG)
    class_discrimination, n_components, perplexity, learning_rate, n_iter, verbose, early_exaggeration, \
        n_iter_without_progress, init, random_state = get_params(CONFIG)

    output = check_outputdir(output)
    sys.stdout = Logger()

    if HAS_ASTROPY:
        X, Y, ID = load_ap(inputs)
    else:
        X, Y, ID = load_csv(inputs)

    """Normalization check"""
    min_max_scaler = preprocessing.MinMaxScaler()
    if normalization == "True":
        X = min_max_scaler.fit_transform(X)

    n_dimensions = X.shape[1]
    Cmap = []
    if class_discrimination == "Boolean":
        for i in Y:
            if i == True:
                Cmap.append("r")
            elif i == False:
                Cmap.append("b")
    elif class_discrimination == "Number":
        Identifiers = np.unique(Y)
        n_classes = len(Identifiers)
        Colors = normalize_list(Identifiers)
        for i in Y:
            for j in Identifiers:
                if i == j:
                    idx = list(Identifiers).index(j)
                    Cmap.append(Colors[idx])

    Model = TSNE(n_components=n_components, perplexity=perplexity, n_iter=n_iter, verbose=verbose,
                 learning_rate=learning_rate, n_iter_without_progress=n_iter_without_progress,
                 early_exaggeration=early_exaggeration, init=init, random_state=random_state, n_jobs=-1)

    X_EMBEDDED = Model.fit_transform(X)
    Parameters = Model.get_params()
    KL_DIVERGENCE = Model.kl_divergence_
    Parameters.update({"kl_divergence": KL_DIVERGENCE,
                       "n_classes": n_classes, "n_dimensions": n_dimensions})
    IO = {"Input": inputs, "Output": output,
          "Log file": logfile_path, "Normalization": normalization}

    now2 = datetime.datetime.now()
    """Plots and Saves the Scatter Plot"""
    plt.figure()
    plt.scatter(X_EMBEDDED[:, 0], X_EMBEDDED[:, 1],
                c=Cmap, alpha=0.7, marker=".", s=5)
    plt.savefig(output + os.sep + "Scatter_Plot.png")

    sys.stdout = sys.__stdout__

    with open(output + os.sep + "logfile.log", "r") as G:
        lines = G.readlines()

    """Write the Parameters in the choosed log .txt"""
    with open(output + os.sep + logfile_path, "w") as file:
        file.write("TSNE Starting Time: " + str(now) + "\n\n")
        file.write(
            "*****************   CONFIG FILE COPY   ******************\n\n")
        file.write("[I/O]\n")
        for key, value in IO.items():
            file.write("%s = %s\n" % (key, value))
        file.write("\n")
        file.write("[TSNE parameters]\n")
        for key, value in Parameters.items():
            file.write("%s = %s\n" % (key, value))
        file.write("\n")
        file.write(
            "*****************   END CONFIG FILE COPY   ******************\n\n")
        file.write("*****************   DURATION SUMMARY   ******************\n")
        file.write("TSNE End Time: " + str(now2) + "\n")
        file.write("TSNE running time: --- %s seconds ---" %
                   (time.time() - start_time) + "\n")
        file.write(
            "*****************   END DURATION SUMMARY   ******************\n\n")
        file.write(
            "*****************   TSNE VERBOSE DUMP   ******************\n\n")
        file.writelines(lines)
        file.close()
    os.remove(output + os.sep + "logfile.log")

    """Writes the Output in a .csv"""
    embedded = rfn.merge_arrays((Y[:, None], X_EMBEDDED[:, 0], X_EMBEDDED[:, 1]), flatten=False, usemask=False, asrecarray=False)
    print(X_EMBEDDED[:, 0][0], X_EMBEDDED[:, 1][0])
    header = ["class", "x", "y"]
    with open(output + os.sep + "Output.csv", "wt", newline="") as file:
        writer = csv.writer(file, delimiter=",")
        writer.writerow(i for i in header)
        for j in embedded:
            writer.writerow(j)
